

class ExitException(Exception):
    pass


def count_leading_spaces(s):
    n=0
    for c in s:
        if c!=' ':
            break
        n+=1
    return n
    
def align(s,n):
    if len(s)>n:
        return s[0:n]
    if len(s)<n:
        return s+' '*(n-len(s))
    return s
    
def ctrl(key):
    return ord(key)-ord('A')+1



#!/usr/bin/env python


class Point(object):
    def __init__(self,*args):
        if len(args)==2:
            self.x=args[0]
            self.y=args[1]
        elif len(args)==1:
            if isinstance(args[0],Point):
                self.x=args[0].x
                self.y=args[0].y
            elif isinstance(args[0],tuple):
                self.x=args[0][0]
                self.y=args[0][1]
            else:
                raise TypeError()
                
    def __str__(self):
        return '{},{}'.format(self.x,self.y)
        
    def __lt__(self,p):
        if not isinstance(p,Point):
            p=Point(p)
        if self.y<p.y:
            return True
        if self.y>p.y:
            return False
        return self.x<p.x
        
    def __le__(self,p):
        return (self<p or self==p)
        
    def __gt__(self,p):
        return not (self<p or self==p)
        
    def __ge__(self,p):
        return not (self<p)
        
    def __eq__(self,p):
        if not isinstance(p,Point):
            p=Point(p)
        return self.x==p.x and self.y==p.y
        
    def __ne__(self,p):
        return not (self==p)

    def __iadd__(self,p):
        if not isinstance(p,Point):
            p=Point(p)
        self.x+=p.x
        self.y+=p.y
        return self

    def __isub__(self,p):
        if not isinstance(p,Point):
            p=Point(p)
        self.x-=p.x
        self.y-=p.y
        return self

    def __add__(self,p):
        if not isinstance(p,Point):
            p=Point(p)
        return Point(self.x+p.x,self.y+p.y)
    
    def __sub__(self,p):
        if not isinstance(p,Point):
            p=Point(p)
        return Point(self.x-p.x,self.y-p.y)

class Rect(object):
    def __init__(self,*args):
        if len(args)==2 and isinstance(args[0],Point) and isinstance(args[1],Point):
            self.tl=Point(args[0])
            self.br=Point(args[1])
        elif len(args)==4:
            self.tl=Point(args[0],args[1])
            self.br=Point(args[2],args[3])
        elif len(args)==1 and isinstance(args[0],Rect):
            self.tl=Point(args[0].tl)
            self.br=Point(args[0].br)
        else:
            raise TypeError()
            
    def normalized(self):
        return Rect(min(self.tl.x,self.br.x),min(self.tl.y,self.br.y),
                    max(self.tl.x,self.br.x),max(self.tl.y,self.br.y))
            
    def width(self):
        return self.br.x-self.tl.x
            
    def height(self):
        return self.br.y-self.tl.y
        
    def inflate(self,d):
        if isinstance(d,Point):
            self.tl-=d
            self.br+=d
        else:
            self.tl-=Point(d,d)
            self.br+=Point(d,d)
        return self
            
    def is_point_inside(self,p):
        if not isinstance(p,Point):
            p=Point(p)
        return p.x>=self.tl.x and p.y>=self.tl.y and p.x<self.br.x and p.y<self.br.y

def unit_test():
    p1=Point(1,2)
    p2=Point((1,2))
    p3=Point(p1)
    s1=str(p1)
    s2=str(p2)
    s3=str(p3)
                
if __name__=='__main__':
    unit_test()
    


from ptypes import Point
import re
import config

def normalize(row,tabsize,tabcount):
    row=row.replace('\n','')
    tabcount[0]+=row.count('\t')
    row=row.replace('\t',' '*tabsize)
    return row

class Document:
    def __init__(self,filename=''):
        self.clear()
        self.filename=filename
        self.tabs_replaced=False
        
    def clear(self):
        self.filename=''
        self.text=['']
        self.valid=False
        self.undos=[]
        self.undoing=False
        self.modified=False
        
    def load(self,filename):
        try:
            self.tabs_replaced=False
            f=open(filename,'r')
            self.text=f.readlines()
            f.close()
            if len(self.text)==0:
                self.text=['']
            else:
                if self.text[-1].endswith('\n'):
                    self.text.append('')
            tabsize=config.getint('tabsize')
            tabcount=[0]
            self.text=[normalize(row,tabsize,tabcount) for row in self.text]
            if tabcount[0]>0:
                self.tabs_replaced=True
            self.filename=filename
            self.invalidate()
        except IOError:
            pass
            #message_box('Failed to open {}'.format(filename))
            
    def saveas(self,path):
        self.filename=path
        return self.save()
        
    def save(self):
        if len(self.filename)==0:
            return False
        f=open(self.filename,'w')
        f.write('\n'.join(self.text))
        f.close()
        self.modified=False
        return True

    def rows_count(self):
        return len(self.text)
        
    def get_row(self,index):
        if index<0 or index>=self.rows_count():
            return ''
        return self.text[index]
        
    def find_in_row(self,cursor,text,case,regex,word):
        if cursor.y<0 or cursor.y>=self.rows_count():
            return -1
        row=self.get_row(cursor.y)
        p=-1
        if not case:
            row=row.lower()
            text=text.lower()
        if word:
            text='(^|\W){}($|\W)'.format(text)
            regex=True
        if regex:
            pat=re.compile(text)
            m=pat.search(row,cursor.x+1)
            if m:
                return 1+m.start()
            return -1
        else:
            p=row.find(text,cursor.x+1)
            return p
        
    def join_next_row(self,row_index):
        if row_index>=0 and row_index<(self.rows_count()-1):
            row=self.get_row(row_index)
            next=self.get_row(row_index+1)
            del self.text[row_index+1]
            self.text[row_index]=row+next
            if not self.undoing:
                self.undos.append([self.new_line,Point(len(row),row_index)])
            self.modified=True
        return Point(0,0)
            
    def insert_block(self,text,cursor):
        if cursor.y<0 or cursor.y>=self.rows_count():
            return False
        row=self.get_row(cursor.y)
        if cursor.x<=len(row):
            if not self.undoing:
                self.undos.append([self.delete_block,cursor.y,cursor.x,cursor.x+len(text)])
            row=row[0:cursor.x]+text+row[cursor.x:]
            self.text[cursor.y]=row
        return Point(len(text),0)

    def delete_block(self,row_index,x0,x1):
        if row_index>=0 and row_index<self.rows_count():
            row=self.get_row(row_index)
            if x1<x0:
                x1=len(row)
            if not self.undoing:
                self.undos.append([self.insert_block,row[x0:x1],Point(x0,row_index)])
            self.text[row_index]=row[0:x0]+row[x1:]
            self.modified=True
        return Point(0,0)
            
    def insert_row(self,row_index,text=''):
        if row_index>=0 and row_index<self.rows_count():
            if not self.undoing:
                self.undos.append([self.delete_row,row_index])
            self.text.insert(row_index,text)
        return Point(0,0)
            
    def delete_row(self,row_index):
        if row_index>=0 and row_index<self.rows_count():
            if not self.undoing:
                self.undos.append([self.insert_row,row_index,self.get_row(row_index)])
            del self.text[row_index]
            self.modified=True
        return Point(0,0)

    def word_right(self,cursor):
        if cursor.y<0 or cursor.y>=self.rows_count():
            return cursor
        row=self.get_row(cursor.y)
        x=cursor.x
        space_found=False
        while x<len(row):
            if not row[x].isalnum():
                space_found=True
            if row[x].isalnum() and space_found:
                break
            x=x+1
        return Point(x,cursor.y)

    def word_left(self,cursor):
        if cursor.y<0 or cursor.y>=self.rows_count():
            return cursor
        row=self.get_row(cursor.y)
        x=cursor.x
        while x>0:
            x-=1
            if x>0 and row[x].isalnum() and not row[x-1].isalnum():
                break
        return Point(x,cursor.y)
        
    def set_cursor(self,cursor):
        if cursor.y<0:
            return Point(0,0)
        if cursor.y>=self.rows_count():
            return Point(0,self.rows_count()-1)
        row=self.get_row(cursor.y)
        x=cursor.x
        if x<0:
            x=0
        if x>len(row):
            x=len(row)
        return Point(x,cursor.y)
        
    def delete_char(self,cursor):
        row=self.text[cursor.y]
        if cursor.x<len(row):
            if not self.undoing:
                self.undos.append([self.add_char,row[cursor.x],Point(cursor),True])
            row=row[0:cursor.x]+row[cursor.x+1:]
            self.text[cursor.y]=row
            self.invalidate()
        else:
            self.join_next_row(cursor.y)
        self.modified=True
        return Point(0,0)
            
    def add_char(self,c,cursor,insert):
        row=self.text[cursor.y]
        if not insert and cursor.x<len(row):
            self.delete_char(cursor)
        row=row[0:cursor.x]+c+row[cursor.x:]
        if not self.undoing:
            self.undos.append([self.delete_char,Point(cursor)])
        self.text[cursor.y]=row
        self.invalidate()
        self.modified=True
        return Point(1,0)
        
    def add_text(self, text, cursor, insert):
        cx=cursor.x
        res=Point(0,0)
        if cursor.x<0 or cursor.y<0 or cursor.y>=self.rows_count():
            return res
        for c in text:
            if ord(c)==10:
                self.new_line(cursor)
                cursor=Point(0,cursor.y+1)
                res=Point(-cx,res.y+1)
            else:
                self.add_char(c,cursor,insert)
                cursor=cursor+Point(1,0)
                res+=(1,0)
        self.modified=True
        return res
    
    def new_line(self, cursor):
        row=self.get_row(cursor.y)
        cur=[row[0:cursor.x],row[cursor.x:]]
        self.text=self.text[0:cursor.y]+cur+self.text[cursor.y+1:]
        if not self.undoing:
            self.undos.append([self.join_next_row,cursor.y])
        self.modified=True
        return Point(-cursor.x,1)

    def invalidate(self):
        self.valid=False
        
    def validate(self):
        self.valid=True

    def start_compound(self):
        if not self.undoing:
            self.undos.append('{')

    def stop_compound(self):
        if not self.undoing:
            self.undos.append('}')

    def undo(self):
        depth=0
        self.undoing=True
        res=Point(0,0)
        while len(self.undos)>0:
            cmd=self.undos[-1]
            del self.undos[-1]
            if isinstance(cmd,str):
                if cmd=='{':
                    depth+=1
                if cmd=='}':
                    depth-=1
            else:
                f=cmd[0]
                res+=f(*cmd[1:])
            if depth==0:
                break
        self.undoing=False
        return res



#!/usr/bin/env python3
import curses
import sys
import utils
import config
from ptypes import Point, Rect
from doc import Document
from view import View
from menus import Menu

class Application:
    def __init__(self):
        self.scr=curses.initscr()
        self.keylog=None
        curses.noecho()
        curses.raw()
        self.scr.keypad(1)
        mx=self.scr.getmaxyx()
        self.width=mx[1]
        self.height=mx[0]
        self.rect=Rect(0,0,self.width,self.height)
        self.menu_bar=Menu('')
        self.shortcuts={}
        curses.start_color()
        curses.use_default_colors()
        curses.init_pair(1,config.getint('fg1',curses.COLOR_YELLOW),
                           config.getint('bg1',curses.COLOR_BLUE))
        curses.init_pair(2,config.getint('fg2',curses.COLOR_WHITE),
                           config.getint('bg2',curses.COLOR_GREEN))
        curses.init_pair(3,config.getint('fg3',curses.COLOR_BLACK),
                           config.getint('bg3',curses.COLOR_WHITE))
        curses.init_pair(4,config.getint('fg4',curses.COLOR_BLACK),
                           config.getint('bg4',curses.COLOR_CYAN))
        curses.init_pair(5,config.getint('fg5',curses.COLOR_YELLOW),
                           config.getint('bg5',curses.COLOR_BLUE))
        curses.init_pair(6,config.getint('fg6',curses.COLOR_YELLOW),
                           config.getint('bg6',curses.COLOR_BLUE))
        curses.init_pair(7,config.getint('fg7',curses.COLOR_WHITE),
                           config.getint('bg7',curses.COLOR_RED))
        sys.stdout.write('\033]12;yellow\007')
        
    def set_menu(self,bar):
        self.menu_bar=bar
        if len(bar.items)>0:
            self.shortcuts['KEY_F(10)']=bar.items[0]

    def move(self,pos):
        if not isinstance(pos,Point):
            pos=Point(pos)
        if self.rect.is_point_inside(pos):
            self.scr.move(pos.y,pos.x)
            return True
        return False
        
    def fill_rect(self,rect,c,clr):
        self.fill(rect.tl.x,rect.tl.y,rect.width(),rect.height(),c,clr)
        
    def fill(self,x0,y0,w,h,c,clr):
        for y in range(y0,y0+h):
            self.move((x0,y))
            for x in range(x0,x0+w):
                self.scr.addch(c,curses.color_pair(clr))
        
    def write(self,text,clr,attr=curses.A_BOLD):
        clr=curses.color_pair(clr)|attr
        if isinstance(text,str):
            for i in range(0,len(text)):
                c=text[i]
                self.scr.addch(c,clr)
        else:
            self.scr.addch(text,clr)
            
    def draw_frame(self,rect,color):
        self.move(rect.tl)
        self.write(curses.ACS_ULCORNER,color)
        for i in range(0,rect.width()-2):
            self.write(curses.ACS_HLINE,color)
        self.write(curses.ACS_URCORNER,color)
        for y in range(rect.tl.y+1,rect.br.y):
            self.move(Point(rect.tl.x,y))
            self.write(curses.ACS_VLINE,color)
            self.move(Point(rect.br.x-1,y))
            self.write(curses.ACS_VLINE,color)
        self.move(Point(rect.tl.x,rect.br.y-1))
        self.write(curses.ACS_LLCORNER,color)
        for i in range(0,rect.width()-2):
            self.write(curses.ACS_HLINE,color)
        self.write(curses.ACS_LRCORNER,color)
        
    def refresh(self):
        self.scr.refresh()
        
    def getkey(self):
        try:
            self.scr.nodelay(False)
            key=self.scr.getkey()
            if len(key)==1 and ord(key[0])==27:
                self.scr.nodelay(True)
                key="Alt+"+self.scr.getkey()
                self.scr.nodelay(False)
        except curses.error:
            key='ESC'
        if key == 'KEY_F(24)': # For debugging purposes
            self.keylog=open('/tmp/key.log','w')
        if not self.keylog is None:
            self.keylog.write('{}   {}\n'.format(key,hex(ord(key[0]))))
            self.keylog.flush()
        return key
        
    def close(self):
        curses.nocbreak()
        self.scr.keypad(0)
        curses.echo()
        curses.endwin()
        self.scr=None
        
    def draw_menu(self):
        color=4
        self.move((0,0))
        self.write(' '*self.width,color)
        self.move((1,0))
        pos=Point(2,1)
        for item in self.menu_bar.items:
            title=item.title
            item.pos=Point(pos)
            pos+=Point(len(title)-title.count('&')+3,0)
            self.write('[',color)
            rev=False
            for c in title:
                if c=='&':
                    rev=True
                else:
                    attr=curses.A_BOLD
                    if rev:
                        attr=curses.A_REVERSE
                        rev=False
                        self.shortcuts['Alt+'+c.upper()]=item
                        self.shortcuts['Alt+'+c.lower()]=item
                    self.write(c,color,attr)
            self.write('] ',color)

def message_box(text):
    pass

def fill_menu(menu,desc):
    for item in desc:
        title=item[0]
        action=item[1]
        if isinstance(action,list):
            m=Menu(title)
            fill_menu(m,action)
            menu.add_item(m)
        else:
            menu.add_item(title,action)
            

def create_menu(view):
    desc=[ ('&File',[ ('&New     Ctrl+N',view.on_file_new),
                      ('&Open    Ctrl+O',view.on_file_open),
                      ('&Save    Ctrl+S',view.on_file_save),
                      ('Save &As',view.on_file_save_as),
                      ('&Exit    Ctrl+Q',view.on_file_exit),
                    ]),
           ('&Edit',[ ('&Copy          Ctrl+C',view.on_copy),
                      ('C&ut           Ctrl+X',view.on_cut),
                      ('&Paste         Ctrl+V',view.on_paste),
                      ('&Find          Ctrl+F',view.on_find_replace),
                      ('&Record Macro  Ctrl+R',view.toggle_macro_record),
                      ('P&lay Macro    Ctrl+P',view.play_macro),
                    ]),
           ('&Options', [ ('&Colors',view.on_colors),
                          ('&Editor',view.on_cfg_editor),
                    ]),
           ('&Help',[ ('&About',view.on_help_about),
                    ]),
         ]
    bar=Menu('')
    fill_menu(bar,desc)
    return bar

def main():
    app=Application()
    filename=''
    if len(sys.argv)>1:
        filename=sys.argv[1]
    doc=Document(filename)
    doc.load(filename)
    view=View(app,doc)
    app.set_menu(create_menu(view))
    app.draw_menu()
    view.render()
    try:
        while view.process_input():
            app.draw_menu()
            view.render()
    except utils.ExitException:
        pass
    
    app.close()


if __name__=='__main__':
    main()


import sys
import atexit

impl=None
copy_func=None
paste_func=None

class qt_Clipboard:
    def __init__(self):
        from PyQt4 import QtGui
        self.app=QtGui.QApplication(sys.argv)
        self.clipboard=self.app.clipboard()
        global copy_func
        global paste_func
        copy_func=self.copy
        paste_func=self.paste
  
    def copy(self,text):
        self.clipboard.setText(text)
    
    def paste(self):
        return self.clipboard.text()
      
    def close(self):
        self.app.exit()


def cleanup():
    if not impl is None:
        impl.close()

atexit.register(cleanup)

def init_qt():
    global impl
    try:
        impl=qt_Clipboard()
    except ImportError:
        pass

def copy(text):
    if not copy_func is None:
        copy_func(text)

def paste():
    if paste_func is None:
        return ''
    return paste_func()

if impl is None:
    init_qt()




import os
import atexit
from configparser import ConfigParser

home=os.environ['HOME']
cfgpath=os.path.join(home,'.yated.ini')
cfg=ConfigParser()
cfg.read(cfgpath)
if not 'config' in cfg:
    cfg['config']={}
section=cfg['config']

def get(name,default=''):
    if not name in section:
        section[name]=default
    return section.get(name)
    
def getint(name,default=0):
    return int(get(name,str(default)))
    
def getbool(name,default=False):
    return (get(name,str(default)) != 'False')
    
def set(name,value):
    section[name]=str(value)
    
def save_cfg():
    with open(cfgpath, 'w') as configfile:
        cfg.write(configfile)

atexit.register(save_cfg)


